#ifndef OALSOUND_H
#define OALSOUND_H

#ifndef NO_OPENAL

#include <vector>
#include <map>

#include "i_sound.h"
#include "s_sound.h"

#ifdef _WIN32
#include <al.h>
#include <alc.h>
#include <alure.h>
#elif defined(__APPLE__)
#include <OpenAL/al.h>
#include <OpenAL/alc.h>
#include <OpenAL/alure.h>
#else
#include <AL/al.h>
#include <AL/alc.h>
#include <AL/alure.h>
#endif

#include "efx.h"


class OpenALSoundStream;

class OpenALSoundRenderer : public SoundRenderer
{
public:
	OpenALSoundRenderer ();
	virtual ~OpenALSoundRenderer ();

	virtual void SetSfxVolume (float volume);
	virtual void SetMusicVolume (float volume);
	virtual SoundHandle LoadSound(BYTE *sfxdata, int length);
	virtual SoundHandle LoadSoundRaw(BYTE *sfxdata, int length, int frequency, int channels, int bits);
	virtual void UnloadSound (SoundHandle sfx);
	virtual unsigned int GetMSLength(SoundHandle sfx);
	virtual unsigned int GetSampleLength(SoundHandle sfx);
	virtual float GetOutputRate();

	// Streaming sounds.
	virtual SoundStream *CreateStream (SoundStreamCallback callback, int buffbytes, int flags, int samplerate, void *userdata);
	virtual SoundStream *OpenStream (const char *filename, int flags, int offset, int length);

	// Starts a sound.
	virtual FISoundChannel *StartSound (SoundHandle sfx, float vol, int pitch, int chanflags, FISoundChannel *reuse_chan);
	virtual FISoundChannel *StartSound3D (SoundHandle sfx, SoundListener *listener, float vol, FRolloffInfo *rolloff, float distscale, int pitch, int priority, const FVector3 &pos, const FVector3 &vel, int channum, int chanflags, FISoundChannel *reuse_chan);

	// Stops a sound channel.
	virtual void StopChannel (FISoundChannel *chan);

	// Returns position of sound on this channel, in samples.
	virtual unsigned int GetPosition(FISoundChannel *chan);

	// Synchronizes following sound startups.
	virtual void Sync (bool sync);

	// Pauses or resumes all sound effect channels.
	virtual void SetSfxPaused (bool paused, int slot);

	// Pauses or resumes *every* channel, including environmental reverb.
	virtual void SetInactive(bool inactive);

	// Updates the volume, separation, and pitch of a sound channel.
	virtual void UpdateSoundParams3D (SoundListener *listener, FISoundChannel *chan, bool areasound, const FVector3 &pos, const FVector3 &vel);

	virtual void UpdateListener (SoundListener *);
	virtual void UpdateSounds ();

	virtual bool IsValid ();
	virtual void PrintStatus ();
	virtual void PrintDriversList ();
	virtual FString GatherStats ();

private:
	// Standard OpenAL 1.1 functions
	static void *openal_handle;
	static LPALCCREATECONTEXT alcCreateContext;
	static LPALCMAKECONTEXTCURRENT alcMakeContextCurrent;
	static LPALCPROCESSCONTEXT alcProcessContext;
	static LPALCSUSPENDCONTEXT alcSuspendContext;
	static LPALCDESTROYCONTEXT alcDestroyContext;
	static LPALCGETCURRENTCONTEXT alcGetCurrentContext;
	static LPALCGETCONTEXTSDEVICE alcGetContextsDevice;
	static LPALCOPENDEVICE alcOpenDevice;
	static LPALCCLOSEDEVICE alcCloseDevice;
	static LPALCGETERROR alcGetError;
	static LPALCISEXTENSIONPRESENT alcIsExtensionPresent;
	static LPALCGETPROCADDRESS alcGetProcAddress;
	static LPALCGETENUMVALUE alcGetEnumValue;
	static LPALCGETSTRING alcGetString;
	static LPALCGETINTEGERV alcGetIntegerv;
	static LPALCCAPTUREOPENDEVICE alcCaptureOpenDevice;
	static LPALCCAPTURECLOSEDEVICE alcCaptureCloseDevice;
	static LPALCCAPTURESTART alcCaptureStart;
	static LPALCCAPTURESTOP alcCaptureStop;
	static LPALCCAPTURESAMPLES alcCaptureSamples;
	static LPALENABLE alEnable;
	static LPALDISABLE alDisable;
	static LPALISENABLED alIsEnabled;
	static LPALGETSTRING alGetString;
	static LPALGETBOOLEANV alGetBooleanv;
	static LPALGETINTEGERV alGetIntegerv;
	static LPALGETFLOATV alGetFloatv;
	static LPALGETDOUBLEV alGetDoublev;
	static LPALGETBOOLEAN alGetBoolean;
	static LPALGETINTEGER alGetInteger;
	static LPALGETFLOAT alGetFloat;
	static LPALGETDOUBLE alGetDouble;
	static LPALGETERROR alGetError;
	static LPALISEXTENSIONPRESENT alIsExtensionPresent;
	static LPALGETPROCADDRESS alGetProcAddress;
	static LPALGETENUMVALUE alGetEnumValue;
	static LPALLISTENERF alListenerf;
	static LPALLISTENER3F alListener3f;
	static LPALLISTENERFV alListenerfv;
	static LPALLISTENERI alListeneri;
	static LPALLISTENER3I alListener3i;
	static LPALLISTENERIV alListeneriv;
	static LPALGETLISTENERF alGetListenerf;
	static LPALGETLISTENER3F alGetListener3f;
	static LPALGETLISTENERFV alGetListenerfv;
	static LPALGETLISTENERI alGetListeneri;
	static LPALGETLISTENER3I alGetListener3i;
	static LPALGETLISTENERIV alGetListeneriv;
	static LPALGENSOURCES alGenSources;
	static LPALDELETESOURCES alDeleteSources;
	static LPALISSOURCE alIsSource;
	static LPALSOURCEF alSourcef;
	static LPALSOURCE3F alSource3f;
	static LPALSOURCEFV alSourcefv;
	static LPALSOURCEI alSourcei;
	static LPALSOURCE3I alSource3i;
	static LPALSOURCEIV alSourceiv;
	static LPALGETSOURCEF alGetSourcef;
	static LPALGETSOURCE3F alGetSource3f;
	static LPALGETSOURCEFV alGetSourcefv;
	static LPALGETSOURCEI alGetSourcei;
	static LPALGETSOURCE3I alGetSource3i;
	static LPALGETSOURCEIV alGetSourceiv;
	static LPALSOURCEPLAYV alSourcePlayv;
	static LPALSOURCESTOPV alSourceStopv;
	static LPALSOURCEREWINDV alSourceRewindv;
	static LPALSOURCEPAUSEV alSourcePausev;
	static LPALSOURCEPLAY alSourcePlay;
	static LPALSOURCESTOP alSourceStop;
	static LPALSOURCEREWIND alSourceRewind;
	static LPALSOURCEPAUSE alSourcePause;
	static LPALSOURCEQUEUEBUFFERS alSourceQueueBuffers;
	static LPALSOURCEUNQUEUEBUFFERS alSourceUnqueueBuffers;
	static LPALGENBUFFERS alGenBuffers;
	static LPALDELETEBUFFERS alDeleteBuffers;
	static LPALISBUFFER alIsBuffer;
	static LPALBUFFERF alBufferf;
	static LPALBUFFER3F alBuffer3f;
	static LPALBUFFERFV alBufferfv;
	static LPALBUFFERI alBufferi;
	static LPALBUFFER3I alBuffer3i;
	static LPALBUFFERIV alBufferiv;
	static LPALGETBUFFERF alGetBufferf;
	static LPALGETBUFFER3F alGetBuffer3f;
	static LPALGETBUFFERFV alGetBufferfv;
	static LPALGETBUFFERI alGetBufferi;
	static LPALGETBUFFER3I alGetBuffer3i;
	static LPALGETBUFFERIV alGetBufferiv;
	static LPALBUFFERDATA alBufferData;
	static LPALDOPPLERFACTOR alDopplerFactor;
	static LPALDOPPLERVELOCITY alDopplerVelocity;
	static LPALDISTANCEMODEL alDistanceModel;
	static LPALSPEEDOFSOUND alSpeedOfSound;
	// ALURE 1.0 functions
	static void *alure_handle;
	static LPALUREGETERRORSTRING alureGetErrorString;
	static LPALUREGETVERSION alureGetVersion;
	static LPALUREGETDEVICENAMES alureGetDeviceNames;
	static LPALUREFREEDEVICENAMES alureFreeDeviceNames;
	static LPALUREINITDEVICE alureInitDevice;
	static LPALURESHUTDOWNDEVICE alureShutdownDevice;
	static LPALUREGETSAMPLEFORMAT alureGetSampleFormat;
	static LPALURESLEEP alureSleep;
	static LPALURECREATEBUFFERFROMFILE alureCreateBufferFromFile;
	static LPALURECREATEBUFFERFROMMEMORY alureCreateBufferFromMemory;
	static LPALUREBUFFERDATAFROMFILE alureBufferDataFromFile;
	static LPALUREBUFFERDATAFROMMEMORY alureBufferDataFromMemory;
	static LPALURECREATESTREAMFROMFILE alureCreateStreamFromFile;
	static LPALURECREATESTREAMFROMMEMORY alureCreateStreamFromMemory;
	static LPALURECREATESTREAMFROMSTATICMEMORY alureCreateStreamFromStaticMemory;
	static LPALURECREATESTREAMFROMCALLBACK alureCreateStreamFromCallback;
	static LPALUREBUFFERDATAFROMSTREAM alureBufferDataFromStream;
	static LPALUREREWINDSTREAM alureRewindStream;
	static LPALUREDESTROYSTREAM alureDestroyStream;
	static LPALUREINSTALLDECODECALLBACKS alureInstallDecodeCallbacks;

	// EFX Extension function pointer variables. Loaded after context creation
	// if EFX is supported. These pointers may be context-dependant, thus can't
	// be static
	// Effect objects
	LPALGENEFFECTS alGenEffects;
	LPALDELETEEFFECTS alDeleteEffects;
	LPALISEFFECT alIsEffect;
	LPALEFFECTI alEffecti;
	LPALEFFECTIV alEffectiv;
	LPALEFFECTF alEffectf;
	LPALEFFECTFV alEffectfv;
	LPALGETEFFECTI alGetEffecti;
	LPALGETEFFECTIV alGetEffectiv;
	LPALGETEFFECTF alGetEffectf;
	LPALGETEFFECTFV alGetEffectfv;
	// Filter objects
	LPALGENFILTERS alGenFilters;
	LPALDELETEFILTERS alDeleteFilters;
	LPALISFILTER alIsFilter;
	LPALFILTERI alFilteri;
	LPALFILTERIV alFilteriv;
	LPALFILTERF alFilterf;
	LPALFILTERFV alFilterfv;
	LPALGETFILTERI alGetFilteri;
	LPALGETFILTERIV alGetFilteriv;
	LPALGETFILTERF alGetFilterf;
	LPALGETFILTERFV alGetFilterfv;
	// Auxiliary slot objects
	LPALGENAUXILIARYEFFECTSLOTS alGenAuxiliaryEffectSlots;
	LPALDELETEAUXILIARYEFFECTSLOTS alDeleteAuxiliaryEffectSlots;
	LPALISAUXILIARYEFFECTSLOT alIsAuxiliaryEffectSlot;
	LPALAUXILIARYEFFECTSLOTI alAuxiliaryEffectSloti;
	LPALAUXILIARYEFFECTSLOTIV alAuxiliaryEffectSlotiv;
	LPALAUXILIARYEFFECTSLOTF alAuxiliaryEffectSlotf;
	LPALAUXILIARYEFFECTSLOTFV alAuxiliaryEffectSlotfv;
	LPALGETAUXILIARYEFFECTSLOTI alGetAuxiliaryEffectSloti;
	LPALGETAUXILIARYEFFECTSLOTIV alGetAuxiliaryEffectSlotiv;
	LPALGETAUXILIARYEFFECTSLOTF alGetAuxiliaryEffectSlotf;
	LPALGETAUXILIARYEFFECTSLOTFV alGetAuxiliaryEffectSlotfv;


	static ALenum checkALError(const char*,unsigned int);

	ALCcontext *Context;

	bool SrcDistanceModel;

	std::vector<ALuint> Sources;

	ALfloat SfxVolume;
	ALfloat MusicVolume;

	int SFXPaused;
	std::vector<ALuint> FreeSfx;
	std::vector<ALuint> PausableSfx;
	std::vector<ALuint> SfxGroup;

	const ReverbContainer *PrevEnvironment;

	typedef std::map<WORD,ALuint> EffectMap;
	ALuint EnvSlot;
	EffectMap EnvEffects;

	ALuint WaterLP;
	float LastWaterLP;

	std::vector<OpenALSoundStream*> Streams;
	friend class OpenALSoundStream;
};

#endif // NO_OPENAL

#endif
